@@ -50,7 +50,7 @@ use alloc::format;
 use alloc::string::ToString;
 use core::cmp::min;
 
-use zune_core::bytestream::{ZByteReader, ZReaderTrait};
+use zune_core::bytestream::{ZByteReaderTrait, ZReader};
 
 use crate::errors::DecodeErrors;
 use crate::huffman::{HuffmanTable, HUFF_LOOKAHEAD};
@@ -87,6 +87,7 @@ macro_rules! decode_huff {
                 // We may think, lets fake zeroes, noo
                 // panic, because Huffman codes are sensitive, probably everything
                 // after this will be corrupt, so no need to continue.
+                // panic!("Bad Huffman code length");
                 return Err(DecodeErrors::Format(format!("Bad Huffman Code 0x{:X}, corrupt JPEG",$symbol)))
             }
 
@@ -103,59 +104,63 @@ macro_rules! decode_huff {
 
 /// A `BitStream` struct, a bit by bit reader with super powers
 ///
+#[rustfmt::skip]
 pub(crate) struct BitStream {
     /// A MSB type buffer that is used for some certain operations
-    pub buffer:           u64,
+    pub buffer:              u64,
     /// A TOP  aligned MSB type buffer that is used to accelerate some operations like
     /// peek_bits and get_bits.
     ///
     /// By top aligned, I mean the top bit (63) represents the top bit in the buffer.
-    aligned_buffer:       u64,
+    aligned_buffer:          u64,
     /// Tell us the bits left the two buffer
-    pub(crate) bits_left: u8,
+    pub(crate) bits_left:    u8,
     /// Did we find a marker(RST/EOF) during decoding?
-    pub marker:           Option<Marker>,
-
-    /// Progressive decoding
-    pub successive_high: u8,
-    pub successive_low:  u8,
-    spec_start:          u8,
-    spec_end:            u8,
-    pub eob_run:         i32,
-    pub overread_by:     usize
+    pub marker:              Option<Marker>,
+    /// An i16 with the bit corresponding to successive_low set to 1, others 0.
+    pub successive_low_mask: i16,
+    spec_start:              u8,
+    spec_end:                u8,
+    pub eob_run:             i32,
+    pub overread_by:         usize,
+    /// True if we have seen end of image marker.
+    /// Don't read anything after that.
+    pub seen_eoi:            bool,
 }
 
 impl BitStream {
     /// Create a new BitStream
+    #[rustfmt::skip]
     pub(crate) const fn new() -> BitStream {
         BitStream {
-            buffer:          0,
-            aligned_buffer:  0,
-            bits_left:       0,
-            marker:          None,
-            successive_high: 0,
-            successive_low:  0,
-            spec_start:      0,
-            spec_end:        0,
-            eob_run:         0,
-            overread_by:     0
+            buffer:              0,
+            aligned_buffer:      0,
+            bits_left:           0,
+            marker:              None,
+            successive_low_mask: 1,
+            spec_start:          0,
+            spec_end:            0,
+            eob_run:             0,
+            overread_by:         0,
+            seen_eoi:            false,
         }
     }
 
     /// Create a new Bitstream for progressive decoding
     #[allow(clippy::redundant_field_names)]
-    pub(crate) fn new_progressive(ah: u8, al: u8, spec_start: u8, spec_end: u8) -> BitStream {
+    #[rustfmt::skip]
+    pub(crate) fn new_progressive(al: u8, spec_start: u8, spec_end: u8) -> BitStream {
         BitStream {
-            buffer:          0,
-            aligned_buffer:  0,
-            bits_left:       0,
-            marker:          None,
-            successive_high: ah,
-            successive_low:  al,
-            spec_start:      spec_start,
-            spec_end:        spec_end,
-            eob_run:         0,
-            overread_by:     0
+            buffer:              0,
+            aligned_buffer:      0,
+            bits_left:           0,
+            marker:              None,
+            successive_low_mask: 1i16 << al,
+            spec_start:          spec_start,
+            spec_end:            spec_end,
+            eob_run:             0,
+            overread_by:         0,
+            seen_eoi:            false,
         }
     }
 
@@ -167,9 +172,9 @@ impl BitStream {
     ///
     /// This function will only refill if `self.count` is less than 32
     #[inline(always)] // to many call sites? ( perf improvement by 4%)
-    fn refill<T>(&mut self, reader: &mut ZByteReader<T>) -> Result<bool, DecodeErrors>
+    pub fn refill<T>(&mut self, reader: &mut ZReader<T>) -> Result<bool, DecodeErrors>
     where
-        T: ZReaderTrait
+        T: ZByteReaderTrait
     {
         /// Macro version of a single byte refill.
         /// Arguments
@@ -179,8 +184,8 @@ impl BitStream {
         macro_rules! refill {
             ($buffer:expr,$byte:expr,$bits_left:expr) => {
                 // read a byte from the stream
-                $byte = u64::from(reader.get_u8());
-                self.overread_by += usize::from(reader.eof());
+                $byte = u64::from(reader.read_u8());
+                self.overread_by += usize::from(reader.eof()?);
                 // append to the buffer
                 // JPEG is a MSB type buffer so that means we append this
                 // to the lower end (0..8) of the buffer and push the rest bits above..
@@ -190,12 +195,12 @@ impl BitStream {
                 // Check for special case  of OxFF, to see if it's a stream or a marker
                 if $byte == 0xff {
                     // read next byte
-                    let mut next_byte = u64::from(reader.get_u8());
+                    let mut next_byte = u64::from(reader.read_u8());
                     // Byte snuffing, if we encounter byte snuff, we skip the byte
                     if next_byte != 0x00 {
                         // skip that byte we read
                         while next_byte == 0xFF {
-                            next_byte = u64::from(reader.get_u8());
+                            next_byte = u64::from(reader.read_u8());
                         }
 
                         if next_byte != 0x00 {
@@ -207,13 +212,15 @@ impl BitStream {
                                 self.aligned_buffer = $buffer << (64 - $bits_left);
                             }
 
-                            self.marker =
-                                Some(Marker::from_u8(next_byte as u8).ok_or_else(|| {
-                                    DecodeErrors::Format(format!(
-                                        "Unknown marker 0xFF{:X}",
-                                        next_byte
-                                    ))
-                                })?);
+                            self.marker = Marker::from_u8(next_byte as u8);
+                            if next_byte == 0xD9 {
+                                // special handling for eoi, fill some bytes,even if its zero,
+                                // removes some panics
+                                self.buffer <<= 8;
+                                self.bits_left += 8;
+                                self.aligned_buffer = self.buffer << (64 - self.bits_left);
+                            }
+
                             return Ok(false);
                         }
                     }
@@ -222,14 +229,30 @@ impl BitStream {
         }
 
         // 32 bits is enough for a decode(16 bits) and receive_extend(max 16 bits)
-        // If we have less than 32 bits we refill
-        if self.bits_left < 32 && self.marker.is_none() {
-            // So before we do anything, check if we have a 0xFF byte
+        if self.bits_left < 32 {
+            if self.marker.is_some() || self.overread_by > 0  || self.seen_eoi{
+                // found a marker, or we are in EOI
+                // also we are in over-reading mode, where we fill it with zeroes
 
-            if reader.has(4) {
+                // fill with zeroes
+                self.buffer <<= 32;
+                self.bits_left += 32;
+                self.aligned_buffer = self.buffer << (64 - self.bits_left);
+                return Ok(true);
+            }
+
+            // we optimize for the case where we don't have 255 in the stream and have 4 bytes left
+            // as it is the common case
+            //
+            // so we always read 4 bytes, if read_fixed_bytes errors out, the cursor is
+            // guaranteed not to advance in case of failure (is this true), so
+            // we revert the read later on (if we have 255), if this fails, we use the normal
+            // byte at a time read
+
+            if let Ok(bytes) = reader.read_fixed_bytes_or_error::<4>() {
                 // we have 4 bytes to spare, read the 4 bytes into a temporary buffer
                 // create buffer
-                let msb_buf = reader.get_u32_be();
+                let msb_buf = u32::from_be_bytes(bytes);
                 // check if we have 0xff
                 if !has_byte(msb_buf, 255) {
                     self.bits_left += 32;
@@ -238,14 +261,13 @@ impl BitStream {
                     self.aligned_buffer = self.buffer << (64 - self.bits_left);
                     return Ok(true);
                 }
-                // not there, rewind the read
-                reader.rewind(4);
+
+                reader.rewind(4)?;
             }
             // This serves two reasons,
             // 1: Make clippy shut up
             // 2: Favour register reuse
             let mut byte;
-
             // 4 refills, if all succeed the stream should contain enough bits to decode a
             // value
             refill!(self.buffer, byte, self.bits_left);
@@ -255,7 +277,6 @@ impl BitStream {
             // Construct an MSB buffer whose top bits are the bitstream we are currently holding.
             self.aligned_buffer = self.buffer << (64 - self.bits_left);
         }
-
         return Ok(true);
     }
     /// Decode the DC coefficient in a MCU block.
@@ -269,10 +290,10 @@ impl BitStream {
     )]
     #[inline(always)]
     fn decode_dc<T>(
-        &mut self, reader: &mut ZByteReader<T>, dc_table: &HuffmanTable, dc_prediction: &mut i32
+        &mut self, reader: &mut ZReader<T>, dc_table: &HuffmanTable, dc_prediction: &mut i32
     ) -> Result<bool, DecodeErrors>
     where
-        T: ZReaderTrait
+        T: ZByteReaderTrait
     {
         let (mut symbol, r);
 
@@ -311,11 +332,11 @@ impl BitStream {
     )]
     #[inline(never)]
     pub fn decode_mcu_block<T>(
-        &mut self, reader: &mut ZByteReader<T>, dc_table: &HuffmanTable, ac_table: &HuffmanTable,
+        &mut self, reader: &mut ZReader<T>, dc_table: &HuffmanTable, ac_table: &HuffmanTable,
         qt_table: &[i32; DCT_BLOCK], block: &mut [i32; 64], dc_prediction: &mut i32
     ) -> Result<(), DecodeErrors>
     where
-        T: ZReaderTrait
+        T: ZByteReaderTrait
     {
         // Get fast AC table as a reference before we enter the hot path
         let ac_lookup = ac_table.ac_lookup.as_ref().unwrap();
@@ -379,6 +400,8 @@ impl BitStream {
     /// Discard the next `N` bits without checking
     #[inline]
     fn drop_bits(&mut self, n: u8) {
+        debug_assert!(self.bits_left >= n);
+        //self.bits_left -= n;
         self.bits_left = self.bits_left.saturating_sub(n);
         self.aligned_buffer <<= n;
     }
@@ -399,22 +422,22 @@ impl BitStream {
     #[allow(clippy::cast_possible_truncation)]
     #[inline]
     pub(crate) fn decode_prog_dc_first<T>(
-        &mut self, reader: &mut ZByteReader<T>, dc_table: &HuffmanTable, block: &mut i16,
+        &mut self, reader: &mut ZReader<T>, dc_table: &HuffmanTable, block: &mut i16,
         dc_prediction: &mut i32
     ) -> Result<(), DecodeErrors>
     where
-        T: ZReaderTrait
+        T: ZByteReaderTrait
     {
         self.decode_dc(reader, dc_table, dc_prediction)?;
-        *block = (*dc_prediction as i16).wrapping_mul(1_i16 << self.successive_low);
+        *block = (*dc_prediction as i16).wrapping_mul(self.successive_low_mask);
         return Ok(());
     }
     #[inline]
     pub(crate) fn decode_prog_dc_refine<T>(
-        &mut self, reader: &mut ZByteReader<T>, block: &mut i16
+        &mut self, reader: &mut ZReader<T>, block: &mut i16
     ) -> Result<(), DecodeErrors>
     where
-        T: ZReaderTrait
+        T: ZByteReaderTrait
     {
         // refinement scan
         if self.bits_left < 1 {
@@ -422,7 +445,7 @@ impl BitStream {
         }
 
         if self.get_bit() == 1 {
-            *block = block.wrapping_add(1 << self.successive_low);
+            *block = block.wrapping_add(self.successive_low_mask);
         }
 
         Ok(())
@@ -436,13 +459,13 @@ impl BitStream {
         return k;
     }
     pub(crate) fn decode_mcu_ac_first<T>(
-        &mut self, reader: &mut ZByteReader<T>, ac_table: &HuffmanTable, block: &mut [i16; 64]
+        &mut self, reader: &mut ZReader<T>, ac_table: &HuffmanTable, block: &mut [i16; 64]
     ) -> Result<bool, DecodeErrors>
     where
-        T: ZReaderTrait
+        T: ZByteReaderTrait
     {
-        let shift = self.successive_low;
         let fast_ac = ac_table.ac_lookup.as_ref().unwrap();
+        let bit = self.successive_low_mask;
 
         let mut k = self.spec_start as usize;
         let (mut symbol, mut r, mut fac);
@@ -458,7 +481,7 @@ impl BitStream {
             if fac != 0 {
                 // fast ac path
                 k += ((fac >> 4) & 15) as usize; // run
-                block[UN_ZIGZAG[min(k, 63)] & 63] = (fac >> 8).wrapping_mul(1 << shift); // value
+                block[UN_ZIGZAG[min(k, 63)] & 63] = (fac >> 8).wrapping_mul(bit); // value
                 self.drop_bits((fac & 15) as u8);
                 k += 1;
             } else {
@@ -471,7 +494,7 @@ impl BitStream {
                     k += r as usize;
                     r = self.get_bits(symbol as u8);
                     symbol = huff_extend(r, symbol);
-                    block[UN_ZIGZAG[k & 63] & 63] = (symbol as i16).wrapping_mul(1 << shift);
+                    block[UN_ZIGZAG[k & 63] & 63] = (symbol as i16).wrapping_mul(bit);
                     k += 1;
                 } else {
                     if r != 15 {
@@ -493,12 +516,12 @@ impl BitStream {
     }
     #[allow(clippy::too_many_lines, clippy::op_ref)]
     pub(crate) fn decode_mcu_ac_refine<T>(
-        &mut self, reader: &mut ZByteReader<T>, table: &HuffmanTable, block: &mut [i16; 64]
+        &mut self, reader: &mut ZReader<T>, table: &HuffmanTable, block: &mut [i16; 64]
     ) -> Result<bool, DecodeErrors>
     where
-        T: ZReaderTrait
+        T: ZByteReaderTrait
     {
-        let bit = (1 << self.successive_low) as i16;
+        let bit = self.successive_low_mask;
 
         let mut k = self.spec_start;
         let (mut symbol, mut r);
@@ -622,9 +645,8 @@ impl BitStream {
         return Ok(true);
     }
 
-    pub fn update_progressive_params(&mut self, ah: u8, al: u8, spec_start: u8, spec_end: u8) {
-        self.successive_high = ah;
-        self.successive_low = al;
+    pub fn update_progressive_params(&mut self, _ah: u8, al: u8, spec_start: u8, spec_end: u8) {
+        self.successive_low_mask = 1i16 << al;
         self.spec_start = spec_start;
         self.spec_end = spec_end;
     }
@@ -650,14 +672,27 @@ fn huff_extend(x: i32, s: i32) -> i32 {
     (x) + ((((x) - (1 << ((s) - 1))) >> 31) & (((-1) << (s)) + 1))
 }
 
-fn has_zero(v: u32) -> bool {
+const fn has_zero(v: u32) -> bool {
     // Retrieved from Stanford bithacks
     // @ https://graphics.stanford.edu/~seander/bithacks.html#ZeroInWord
     return !((((v & 0x7F7F_7F7F) + 0x7F7F_7F7F) | v) | 0x7F7F_7F7F) != 0;
 }
 
-fn has_byte(b: u32, val: u8) -> bool {
+const fn has_byte(b: u32, val: u8) -> bool {
     // Retrieved from Stanford bithacks
     // @ https://graphics.stanford.edu/~seander/bithacks.html#ZeroInWord
-    has_zero(b ^ ((!0_u32 / 255) * u32::from(val)))
+    has_zero(b ^ ((!0_u32 / 255) * (val as u32)))
 }
+
+// mod tests {
+//     use zune_core::bytestream::ZCursor;
+//     use crate::JpegDecoder;
+//
+//     #[test]
+//     fn test_image() {
+//         let img = "/Users/etemesi/Downloads/PHO00008.JPG";
+//         let data = std::fs::read(img).unwrap();
+//         let mut decoder = JpegDecoder::new(ZCursor::new(&data[..]));
+//         decoder.decode().unwrap();
+//     }
+// }
